/* 
   This file implements a kernel space part of the Cygnal 
   (now Silicon Laboratories) C2 interface driver
   It has been written by Wojciech M. Zabolotny (wzab@ise.pw.edu.pl)
   and is published under the GPL (Gnu Public License) version 2.0 or later

   The code is partially based on examples from the book written
   by Alessandro Rubini & Jonathan Corbet - "Linux Device Drivers"

   All the C2 related information was taken from the Cygnal's
   "C2 Interface Specification"

   For hardware level portability, (eg. porting to ETRAX platform?)
   we assume, that we have some macros for setting/reading I/O lines
   used to comunicate with the C2 interface.
 
   Anyway from the software point of wiev we should have independent macros:

   C2CK_ENA; - enables the clock/reset line
   C2CK_DIS; - disables the clock/reset line
   C2CK_SET(int); (0 or 1)
   C2CK_HIGH();
   C2CK_LOW();

   C2D_READ(); - reads the C2D line
   C2D_ENA(); - enables the C2D line
   C2D_DIS(); - disables the C2D line
   C2D_SET(int); (0 or 1)
   C2D_SETL(); - sets low
   C2D_SETH(); - sets high
   C2D_SETL_EN(); - sets high and enables
   C2D_SETH_EN(); - sets low and enables

   Hardware dependent macros are defined in the c2_lpt.h file
*/
   
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
                                                                                
/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
                                                                                
//#include <linux/fs.h>
//#include <linux/mm.h>
#include <linux/config.h>

#include <linux/ioport.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif
                                                                                
#include <linux/ioctl.h>
#include <asm/uaccess.h>  /* for put_user */

static int io_base=0x378;
MODULE_PARM(io_base,"i");
static int major=0; // requested major number
MODULE_PARM(major,"i");

#define DEVICE_NAME "cygnal"
int Major = 0; // assigned major number

static int c2_busy=0;

//Include the hardware dependent macros
#include "c2_lpt.h"
#include "c2_ioctls.h"


void C2CK_low_pulse()
{
  //We disable interrupts for the current CPU - the pulse must not be too long!!!
  //Later we should decide if we need a spinlock there...
  unsigned long flags;
  local_irq_save(flags);
  C2CK_LOW();
  udelay(1); // 1 us should be enough, and safe
  C2CK_HIGH();
  local_irq_restore(flags);
  //printk("<1>C2CK pulse\n");
}

/* The C2CK_reset function sends the low pulse on the C2CK line long enough to 
   reset our device.
*/
void C2CK_reset()
{
  //This time we do not disable interrupts, as it is not a problem
  //if the pulse lasts longer
  C2CK_LOW();
  udelay(40); // it should be at least 20 us, let's do it twice as long
  C2CK_HIGH();
  udelay(3); //we need to wait at least 2us before we do something else
}

void AddressWrite(unsigned char a)
{
  int i;
  C2CK_low_pulse(); //start
  udelay(1);
  C2D_SETH_EN();
  C2CK_low_pulse();
  udelay(1);
  C2CK_low_pulse();
  udelay(1);
  //Now we are sending the address bits
  for (i=0;i<8;i++) {
    if(a & 1)
      C2D_SETH();
    else
      C2D_SETL();
    C2CK_low_pulse();
    a >>= 1;
    udelay(1);
  }
  C2D_DIS();
  C2CK_low_pulse();
  udelay(1);
}

unsigned char AddressRead()
{
  int i;
  unsigned char a=0, mask=1;
  C2CK_low_pulse(); //start
  udelay(1);
  C2D_SETL_EN();
  C2CK_low_pulse();
  udelay(1);
  C2D_SETH();
  C2CK_low_pulse();
  udelay(1);
  C2D_DIS();
  //Now we are receiving the address bits
  for (i=0;i<8;i++) {
    C2CK_low_pulse();
    udelay(1);
    if(C2D_READ())
      a |= mask;
    mask <<=1;
  }
  C2CK_low_pulse();
  udelay(1);
  return a;
}

/* The data write, and read commands are dangerous, as they may lock up 
   the system (if the WAIT will last forever due to a hardware failure)
   that's why we check for the signal pending, and call the schedule function.
*/

int DataWrite(unsigned char l, unsigned long a)
{
  int i, count=0;
  C2CK_low_pulse(); //start
  udelay(1);
  C2D_SETH_EN();
  C2CK_low_pulse();
  udelay(1);
  C2D_SETL();
  C2CK_low_pulse();
  udelay(1);
  //Now we are sending the length bits
  if (l & 1) 
    C2D_SETH();
  else
    C2D_SETL();
  C2CK_low_pulse();
  udelay(1);
  if (l & 2) 
    C2D_SETH();
  else
    C2D_SETL();
  C2CK_low_pulse();
  udelay(1);
  //Finally we send the data bits
  for (i=0;i<8*(l+1);i++) {
    if(a & 1)
      C2D_SETH();
    else
      C2D_SETL();
    C2CK_low_pulse();
    a >>= 1;
    udelay(1);
  }
  C2D_DIS();
  //And now we enter the WAIT state
  //To avoid deadlock, we need to count the WAIT cycles, check for signals,
  //and call schedule, wherever waiting lasts too long
  while(1) {
    C2CK_low_pulse();
    udelay(1);
    if(C2D_READ())
      break; // We've got `1`!!!
    if(++count == 200) { //OK 200 is just an arbitrary value
      if (signal_pending(current)) {
	return -ERESTARTSYS;
      }
      count=0;
      schedule();
    }
  }
  //We've left the loop, so now we need only to generate the stop pulse
  C2CK_low_pulse();
  udelay(1);
  return 0;
}

int DataRead(unsigned char l, unsigned long *a)
{
  int i, count=0;
  unsigned mask=1;
  *a=0;
  C2CK_low_pulse(); //start
  udelay(1);
  C2D_SETL_EN();
  C2CK_low_pulse();
  udelay(1);
  C2CK_low_pulse();
  udelay(1);
  //Now we are sending the length bits
  if (l & 1) 
    C2D_SETH();
  else
    C2D_SETL();
  C2CK_low_pulse();
  udelay(1);
  if (l & 2) 
    C2D_SETH();
  else
    C2D_SETL();
  C2CK_low_pulse();
  udelay(1);
  //And now we enter the WAIT state
  //To avoid deadlock, we need to count the WAIT cycles, check for signals,
  //and call schedule, wherever waiting lasts too long
  C2D_DIS();
  while(1) {
    C2CK_low_pulse();
    udelay(1);
    if(C2D_READ())
      break; // We've got `1`!!!
    if(++count == 200) { //OK 200 is just an arbitrary value
      if (signal_pending(current)) {
	return -ERESTARTSYS;
      }
      count=0;
      schedule();
    }
  }
  //We've left the loop, so now we can read the data bits
  for (i=0;i<8*(l+1);i++) {
    //for (i=0;i<8;i++) {
    C2CK_low_pulse();
    udelay(1);
    if(C2D_READ())
      (*a) |= mask;
    mask <<= 1;
  }
  C2CK_low_pulse();
  udelay(1);
  return 0;
}

/* Here we have the standard functions for the driver */

/* Procedure servicing the ioctl */
int c2_ioctl (struct inode *inode, struct file *filp,
	      unsigned int cmd, unsigned long arg)
{
  long buf;
  unsigned char cbuf;
  int err=0,ret;
  if (_IOC_TYPE(cmd) != C2_IOC_MAGIC) return -ENOTTY;
  if (_IOC_NR(cmd) > C2_IOC_MAXNR) return -ENOTTY;
  if (_IOC_DIR(cmd) & _IOC_READ)
    err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
  else if (_IOC_DIR(cmd) & _IOC_WRITE)
    err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
  if (err) return -EFAULT;
  switch(_IOC_NR(cmd)) {
  case _IOC_NR(C2_IOCRESET):
    C2CK_reset();
    return 0;
  case _IOC_NR(C2_IOCAWRITE):
    ret = __get_user(cbuf, (unsigned char *)arg);
    AddressWrite(cbuf);
    //printk("<1>Zapisalem adres: %d\n",cbuf);
    return ret;
  case _IOC_NR(C2_IOCDWRITE):
    ret = __get_user(buf, (long *)arg);
    return DataWrite((_IOC_SIZE(cmd)-1) & 3, buf);
  case _IOC_NR(C2_IOCAREAD):
    cbuf=AddressRead();
    //printk("<1>Odczytalem adres: %d\n",cbuf);
    ret = __put_user(cbuf, (unsigned char *)arg);
    return ret;
  case _IOC_NR(C2_IOCDREAD):
    ret = DataRead((_IOC_SIZE(cmd)-1) & 3, &buf);
    if(ret) return ret;
    ret = __put_user(buf, (long *)arg);
    return ret;
  default:
    return -ENOTTY;
  }
}

spinlock_t c2_lock = SPIN_LOCK_UNLOCKED;

static int c2_open(struct inode *inode,
		      struct file *file)
{
  spin_lock(&c2_lock);
  if(c2_busy) {
    spin_unlock(&c2_lock);
    return -EBUSY;
  }
  c2_busy++;
  spin_unlock(&c2_lock);
  MOD_INC_USE_COUNT;
  C2CK_HIGH();
  C2CK_ENA();
  return 0;
}
                                                                                
static int c2_release(struct inode *inode,
			 struct file *file)
{
  spin_lock(&c2_lock);
  c2_busy--;
  spin_unlock(&c2_lock);
  MOD_DEC_USE_COUNT;
  //Now we need to set reasonable state of interface lines
  C2D_DIS();
  C2CK_DIS();
  return 0;
}


/* File operations structure - only ioctl, open and release implemented*/                                                                                
struct file_operations Fops = {
  .ioctl=c2_ioctl, /* ioctl */
  .open=c2_open,
  .release=c2_release,
};


/* Initialization */
int c2_init()
{
  /* Register the character device (at least try) */
  Major = register_chrdev(major,
                          DEVICE_NAME,
                          &Fops);
  /* Negative values signify an error */
  if (Major < 0) {
    printk ("<1>%s device failed with %d\n",
            "Sorry, registering the character",
            Major);
    return Major;
  }
  request_region(io_base,0x10,DEVICE_NAME);
  if(major) Major=major; /* The register_chrdev returns "0" if
			    major number was given, and registration
			    is successfull */
  printk ("<1>%s The major device number is %d.\n",
          "Registeration is a success.",
          Major);
  return 0;
}

/* Cleanup when removing the module */
void c2_exit()
{
  int ret;
  release_region(io_base,0x10);
  ret = unregister_chrdev(Major, DEVICE_NAME);
  /* If there's an error, report it */
  if (ret < 0)
    printk("<1>Error in unregister_chrdev: %d\n", ret);
}
                                                                                
module_init(c2_init);
module_exit(c2_exit);
MODULE_LICENSE("GPL");
